#%%
from dcctk.corpusReader import PlainTextReader
from dcctk.concordancer import Concordancer
from dcctk.corpus import TextBasedCorpus
from dcctk.embeddings import AnchiBert, semantic_sort

c = Concordancer(PlainTextReader("data/").corpus)

#%%
# c = TextBasedCorpus(PlainTextReader().corpus)
# c.list_files('三國')
# c.get_meta_by_path('03/三國志_蜀書一.txt')
# c.get_text('03/三國志_蜀書一.txt', as_str=True)
texts = c.get_texts('三國志', texts_as_str=False, sents_as_str=True)
texts_str = c.get_texts('三國志', texts_as_str=True, sents_as_str=True)

#%%
cql = '''
"法" "院"
'''.strip()
results = list(c.cql_search(cql, left=10, right=10))
print(len(results))
x = results[0]
#%%
x.get_kwic()
x.get_timestep()

#%%
# Sort Concord
results_sorted = sorted(results, key=lambda x: x.get_timestep())
# results[0].get_timestep()

#%%
emb = AnchiBert()

#%%
base_sent, base_tk = results[17].get_kwic()
print(base_sent)
print(base_tk)
sem_sort_by_tk = semantic_sort(emb, results[:200], base_sent, base_tk)
sem_sort_by_sent = semantic_sort(emb, results[:200], base_sent)

#%%
import gdown


gdown.download('https://drive.google.com/uc?id=1uMlNJzilEhSigIcfjTjPdYOZL9IQfHNK', output="AnchiBERT.zip")

#%%
gdown.extractall("AnchiBERT.zip")

# %%
import cqls
queries = cqls.parse(cql, default_attr=c._cql_default_attr,max_quant=6)

x = c._search_keyword(queries[0][0])
    # for result in c._kwic(keywords=query, left=5, right=5):
    #     results.append(result)
# %%
import math

query = queries[0]

best_search_loc = (0, None, math.inf)
for i, keyword in enumerate(query):
    results = c._search_keyword(keyword)
    num_of_matched = len(results)
    if num_of_matched == 0: 
        pass
    elif num_of_matched < best_search_loc[-1]:
        best_search_loc = (i, results, num_of_matched)
results = best_search_loc[1]


#%%
from dcctk.UtilsConcord import queryMatchToken

keyword_anchor = {
    'length': len(query),
    'seed_idx': best_search_loc[0]
}
keywords = query

matched_results = []
for idx in results:
    # Get all possible matching keywords from corpus
    candidates = c._get_keywords(keyword_anchor, *idx)
    if len(candidates) != len(keywords): 
        continue
    # Check every token in keywords
    matched_num = 0
    for w_k, w_c in zip(keywords, candidates):
        if queryMatchToken(queryTerm=w_k, corpToken=w_c):
            matched_num += 1
    if matched_num == len(keywords):
        first_keyword_idx = idx[2] - keyword_anchor['seed_idx']
        matched_results.append( [idx[0], idx[1], first_keyword_idx] )
# %%



import numpy as np
import pandas as pd
from tqdm.auto import trange
from .embeddings import AnchiBert
from .UtilsGeneral import stringify_obj, flatten


class SenseAnalysis:

    def __init__(self, concord_lines:list, bert_model:AnchiBert, is_traditional=True, token_idx:tuple=None, hover_df:pd.DataFrame=None):
        
        self.model = bert_model
        self.concords = concord_lines
        self.concords_len = len(self.concords)        
        self.embeddings = np.empty((self.concords_len, 768))
        self._get_embeddings(is_traditional, token_idx)
        self.embeddings_plotting = None

        # Plotting info
        self.hover_df = hover_df
        if hover_df is None:
            hover_df = []
            for x in self.concords:
                d = {
                    'keyword': x.data['keyword'],
                    'kwic': x.data['left'] + '{' + x.data['keyword'] + '}' + x.data['right'],
                    'timestep': str(x.get_timestep()),
                }
                for k, v in flatten(x.data['meta'], parent_key='m').items():
                    d[k] = stringify_obj(v)
                hover_df.append(d)
            self.hover_df = pd.DataFrame(hover_df)
    

    def plot_embeddings(self, interactive=True, labels:str=None, **keywords):
        import umap.plot
        if self.embeddings_plotting is None:
            self.embeddings_plotting = self.embeddings_dim_reduce(dim=2, method="umap")
        labels = self.hover_df[labels] if labels is not None else None

        if interactive:
            from bokeh.plotting import show    
            f = umap.plot.interactive(self.embeddings_plotting, 
                            labels=labels, 
                            hover_data=self.hover_df, 
                            **keywords)
            show(f)
        else:
            umap.plot.points(self.embeddings_plotting, labels=labels, **keywords)


    def embeddings_dim_reduce(self, dim=2, method="umap"):
        if method.lower() == "umap":
            import umap
            emb = umap.UMAP(n_components=dim, metric='cosine').fit(self.embeddings)
        if method.lower() == "pca":
            from sklearn.decomposition import PCA
            emb = PCA(n_components=dim).fit_transform(self.embeddings)
        return emb


    def _get_embeddings(self, is_traditional=True, token_idx:tuple=None):
        print('Computing embeddings...')
        for i in trange(self.concords_len):
            sent, tk_idx = self.concords[i].get_kwic()
            if token_idx is not None:
                tk_idx = token_idx
            self.embeddings[i] = self.model.encode_sentence(sent, *tk_idx, is_traditional=is_traditional)




import json
from pathlib import Path
from .UtilsTextProcess import *


class PlainTextReader:

    def __init__(self, dir_path="data/", ts_meta_filename="time.yaml", text_meta_filename="text_meta.yaml", ts_meta_loader=None, text_meta_loader=None, plain_text_reader=read_text_as_sentences):
        # Attributes
        self.corp_path = Path(dir_path)
        self.corpus = []
        self.text_meta = {}
        self.timestep_meta = {}
        self.plain_text_reader = plain_text_reader
        self.timestep_meta = self.get_meta(self.corp_path / ts_meta_filename, custom_loader=ts_meta_loader)
        self.text_meta = self.get_meta(self.corp_path / text_meta_filename, custom_loader=text_meta_loader)
        self.read_corpus()


    def read_corpus(self):
        for dir_ in sorted(self.corp_path.iterdir()):
            if not dir_.is_dir(): continue
            
            # Construct corpus
            meta = {}
            for k, v in self.timestep_meta.get(dir_.stem, {}).items():
                meta[k] = v
            corpus = {
                'id': dir_.stem,
                'm': meta,
                'text': []
            }

            # Read text
            for fp in dir_.glob("*.txt"):
                text_id, text_meta, text_content = self.read_text(fp)
                text = {
                    'id': text_id,
                    'm': text_meta,
                    'c': list(text_content)  # A list of sentences
                }
                corpus["text"].append(text)

            self.corpus.append(corpus)


    def read_text(self, fp):
        meta = {}
        fp_key = f"{fp.parent.stem}/{fp.name}"
        for k, v in self.text_meta.get(fp_key, {}).items():
            meta[k] = v

        return fp_key, meta, self.plain_text_reader(fp)



    def get_meta(self, fp, custom_loader=None):
        with open(fp, encoding="utf-8") as f:
            if custom_loader:
                return custom_loader()
            elif fp.name.endswith('yaml'):
                import yaml
                return yaml.load(f, Loader=yaml.FullLoader)
            elif fp.name.endswith('json'):
                return json.load(f)




# if __name__ == '__main__':
#     for sent in read_text_as_sentences("data/01/儀禮_喪服.txt"):
#         print(sent)


import re
from tqdm.auto import trange


class TextBasedCorpus:
    """Corpus object for text-based (text as unit) analysis
    """

    def __init__(self, corpus):
        self.corpus = corpus
        self.path_index = {}
        self.index_path()


    def get_texts(self, pattern, texts_as_str=False, sents_as_str=True):
        texts = {}
        for id in self._list_pattern(pattern):
            text = self.get_text(id, as_str=False)
            if text is None: continue
            if sents_as_str:
                texts[id] = '\n'.join(text)
            else:
                texts[id] = text
        if texts_as_str: 
            return '\n'.join(texts.values())
        return texts
            

    def get_text(self, id, as_str=False):
        idx = self.path_index.get(id, None)
        if idx is None or isinstance(idx, int): 
            return None
        i, j = idx
        text = self.corpus[i]['text'][j].get('c', [])
        if as_str:
            text = '\n'.join(text)
        return text


    def get_meta_by_path(self, id):
        idx = self.path_index.get(id, None)
        if idx is None:
            return {}
        if isinstance(idx, int):
            return self.corpus[idx].get('m', {})
        if isinstance(idx, tuple):
            i, j = idx
            return self.corpus[i]['text'][j].get('m', {})
        return {}


    def list_files(self, pattern, generator=False):
        if generator:
            return self._list_pattern(pattern)
        return list(self._list_pattern(pattern))


    def _list_pattern(self, pattern):
        pattern = re.compile(pattern)
        for k in self.path_index.keys():
            if pattern.search(k):
                yield k


    def index_path(self):
        print("Indexing corpus for text retrival...")
        for i in trange(len(self.corpus)):
            self.path_index[self.corpus[i]['id']] = i
            for j, text in enumerate(self.corpus[i]['text']):
                self.path_index[text['id']] = (i, j)



class IndexedCorpus(TextBasedCorpus):
    """Corpus object for fast concordance search
    """

    def __init__(self, corpus) -> None:
        TextBasedCorpus.__init__(self, corpus)
        self.index = {}
        self.index_corpus()
    

    def get_meta(self, subcorp_idx, text_idx=None, keys:list=None, include_id=True):
        if text_idx is None:
            meta = self.corpus[subcorp_idx]['m']
            if include_id:
                meta['id'] = self.corpus[subcorp_idx]['id']
        else:
            meta = self.corpus[subcorp_idx]['text'][text_idx]['m']
            if include_id:
                meta['id'] = self.corpus[subcorp_idx]['text'][text_idx]['id']
        if keys:
            keys.append('id')
            return { k:meta[k] for k in keys if k in meta }
        return meta


    def index_corpus(self):
        print("Indexing corpus for concordance search...")
        for i in trange(len(self.corpus)):
            for j, text in enumerate(self.corpus[i]['text']):
                for k, sent in enumerate(text['c']):
                    for l, char in enumerate(sent):
                        if char not in self.index:
                            self.index[char] = []
                        self.index[char].append( (i, j, k, l) )




from scipy import spatial

def cossim(x, y) -> float:
    """[summary]

    Parameters
    ----------
    x : [type]
        [description]
    y : [type]
        [description]

    Returns
    -------
    float
        [description]
    """
    return 1 - spatial.distance.cosine(x, y)



import torch
import opencc
from tqdm.auto import trange
from typing import Callable
from pathlib import Path
from transformers import BertTokenizer, BertConfig, BertModel
from .UtilsStats import cossim


class AnchiBert:

    def __init__(self, model_path=None):
        if model_path is None:
            model_path = 'AnchiBERT/'
            if not Path(model_path).is_dir():
                _download_bert_model()

        # Bert initialization
        print(f'Loading AnchiBERT model from {model_path} ...')
        config = BertConfig.from_pretrained(model_path)
        self.tokenizer = BertTokenizer.from_pretrained(model_path)
        self.model = BertModel.from_pretrained(model_path,config=config)

        # Traditional to simplified chinese
        self.t2s = opencc.OpenCC('t2s.json')


    def encode_sentence(self, sent:str, idx_from:int=None, idx_to:int=None, is_traditional=True):
        """Encode a raw sentence as a vector

        Parameters
        ----------
        sent : str
            The sentence to encode.
        idx_from : int, optional
            Begining index of a token in the :code:`sent`, 
            by default None.
        idx_to : int, optional
            Ending index of a token in the :code:`sent`, 
            by default None. If :code:`idx_from` or 
            :code:`idx_to` is None, the sentence would be
            encoded by summing all tokens' vectors from the
            last hidden state in the Bert model. Otherwise,
            the subset of tokens in :code:`sent` as specified 
            by :code:`(idx_from, idx_to)` are used.
        is_traditional : bool, optional
            [description], by default True

        Returns
        -------
        numpy.ndarray
            A :code:`(768,)` numpy.ndarray array.
        """
        if is_traditional: sent = self.t2s.convert(sent)
        input = self.tokenizer(sent, return_tensors="pt")
        last_hidden_state = self.model(**input)[0]

        # Sentence vector (sum all tokens together)
        if idx_from is None or idx_to is None:
            return torch.sum(last_hidden_state[0, :, :], dim=0).detach().numpy()
        # Token vector (sum subset of tokens together)
        return torch.sum(last_hidden_state[0, (idx_from+1):(idx_to+2), :], dim=0).detach().numpy()



def semantic_sort(m:AnchiBert, concord_lines:list, base_sent:str=None, base_tk:tuple=None, is_traditional=True, simil_func:Callable=cossim):
    """Sort concordance lines based on semantic similarity of
    vectors from AnchiBert model

    Parameters
    ----------
    m : AnchiBert
        Initialized AnchiBert model from 
        :class:`embeddings.AnchiBert`.
    concord_lines : list
        A list of :class:`concordancer.ConcordLine` obtained 
        from :meth:`concordancer.Concordancer.cql_search`.
    base_sent : str, optional
        The sentence used to compare with, by default None,
        which takes the first element of :code:`concord_lines`
        as the base sentence.
    base_tk : tuple, optional
        A tuple of length 2, specifying the position 
        :code:`(idx_from, idx_to)` of the keywords in
        :code:`base_sent`, by default None. If None, the
        vector is computed from summing up all token vectors in the sentence. Otherwise, only keywords' vectors are used
        to derive the vector.
    is_traditional : bool, optional
        Whether to convert traditional into simplified Chinese
        before feeding the input to AnchiBert, by default True.
        This parameter is passed to 
        :meth:`embeddings.AnchiBert.encode_sentence`.
    simil_func : Callable, optional
        Function for computing similarity between two vectors,
        by default :func:`UtilsStats.cossim`. The function 
        should take the input vectors as its two arguments and
        returns a float.

    Returns
    -------
    list
        A list of 2-tuples :code:`(<ConcordLine>, <float>)`, 
        with the second elements being the similarity scores
        to :code:`base_sent (base_tk)`.
    """
    if base_sent is None:
        base_sent, base_tk = concord_lines[0].get_kwic()
    
    if base_tk is not None:
        compare_base = m.encode_sentence(base_sent, *base_tk, is_traditional=is_traditional)
    else:
        compare_base = m.encode_sentence(base_sent, is_traditional=is_traditional)

    results = []
    for i in trange(len(concord_lines)):
        result = concord_lines[i]
        sent, tk_idx = result.get_kwic()
        if base_tk is not None:
            vec = m.encode_sentence(sent, *tk_idx, is_traditional=is_traditional)
        else:
            vec = m.encode_sentence(sent, is_traditional=is_traditional)
        
        results.append( (result, simil_func(compare_base, vec)) )

    return sorted(results, key=lambda x: x[1], reverse=True)



def _download_bert_model():
    print("Downloading AnchiBert model ...")
    import gdown
    gdown.download('https://drive.google.com/uc?id=1uMlNJzilEhSigIcfjTjPdYOZL9IQfHNK', quiet=False, output="AnchiBERT.zip")
    gdown.extractall("AnchiBERT.zip")



def read_text_as_sentences(fp):
    with open(fp, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line == "": continue
            yield line
            # if "。" in line:
            #     lines = line.split("。")
            #     last_idx = len(lines) - 1
            #     for i, line in enumerate(lines):
            #         if i != last_idx: line += "。"
            #         if line == "": continue
            #         yield line
            # else:
                # yield line



from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
      long_description = f.read().strip()


with open("VERSION") as f:
      v = int(f.read().strip()) + 1

setup(name='dcctk',
      version=f'0.0.{v}',
      description="Diachronic Character-based Corpus toolkit",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/liao961120/dcctk',
      author='Yongfu Liao',
      author_email='liao961120@github.com',
      license='MIT',
      packages=['dcctk'],
      install_requires=['pyyaml', 'cqls', 'gdown>=3.10.2', 'opencc', 'transformers', 'scipy'],
      # tests_require=['deepdiff'],
      zip_safe=False)


with open("VERSION", "w") as f:
      f.write(str(v))
